"use strict";

var _interopRequireDefault = require("@babel/runtime/helpers/interopRequireDefault");

exports.__esModule = true;
exports.InputContainer = void 0;

var _immer = _interopRequireDefault(require("immer"));

var _react = _interopRequireWildcard(require("react"));

var _rxjs = require("rxjs");

var _operators = require("rxjs/operators");

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

class InputContainer extends _react.PureComponent {
  constructor(props) {
    super(props);
    this.destroy$ = new _rxjs.Subject();
    this.state = {
      status: "SUCCESS"
    };
    this.initializeControl.bind(this);
  }

  componentDidMount() {
    const control = this.props.control;
    control && this.initializeControl(control);
  }

  initializeControl(control) {
    control.statusChanges.pipe((0, _operators.takeUntil)(this.destroy$), (0, _operators.startWith)(control.status), (0, _operators.map)(statusType => {
      if (statusType === "INVALID") {
        return {
          status: "ERROR",
          error: control.errors
        };
      } else if (statusType === "PENDING") {
        return {
          status: "PROCESSING",
          data: {}
        };
      }

      return {
        status: "SUCCESS",
        data: {}
      };
    }), (0, _operators.tap)(status => {
      this.setState((0, _immer.default)(draft => {
        draft.data = status.data;
        draft.error = status.error;
        draft.status = status.status;
      }));
    })).subscribe();
  }

  componentDidUpdate(prevProps) {
    let newControl = this.props.control;

    if (newControl && prevProps.control !== newControl) {
      this.initializeControl(newControl);
    }
  }

  componentWillUnmount() {
    this.destroy$.next();
    this.destroy$.complete();
  }

  render() {
    const control = this.props.control;
    const showErrorsFn = this.props.showErrorsFn ? this.props.showErrorsFn : () => true;
    const formLoadStatus = this.props.formLoadStatus ? this.props.formLoadStatus : "SUCCESS";
    const showErrors = control ? showErrorsFn(control) : false;
    const data = this.props.data ? this.props.data : {};
    const dataLoadStatus = data.load ? data.load : {
      status: "SUCCESS"
    };
    const renderLabel = this.props.renderLabel;
    const labelTemplate = renderLabel && renderLabel({
      showErrors,
      validationStatus: this.state.status
    });
    const renderInput = this.props.renderInput;
    const inputTemplate = control && formLoadStatus === "SUCCESS" && dataLoadStatus.status === "SUCCESS" && renderInput({
      control,
      data: dataLoadStatus.data,
      validationStatus: this.state.status,
      showErrors
    });
    const renderProcessing = this.props.renderProcessing;
    const processingTemplate = (formLoadStatus === "PROCESSING" || dataLoadStatus.status === "PROCESSING") && renderProcessing && renderProcessing();
    const renderValidationErrors = this.props.renderValidationErrors;
    const validationErrorsTemplate = control && formLoadStatus === "SUCCESS" && dataLoadStatus.status === "SUCCESS" && showErrors && renderValidationErrors && renderValidationErrors({
      errors: this.state.error,
      control
    });
    const renderReload = data.renderReload;
    const reloadTemplate = dataLoadStatus.status === "ERROR" && renderReload && renderReload(dataLoadStatus.error);
    return /*#__PURE__*/_react.default.createElement(_react.default.Fragment, null, formLoadStatus !== "ERROR" && /*#__PURE__*/_react.default.createElement(_react.default.Fragment, null, labelTemplate, inputTemplate, processingTemplate, validationErrorsTemplate, reloadTemplate), formLoadStatus === "SUCCESS" && dataLoadStatus.status === "SUCCESS" && !this.props.control && /*#__PURE__*/_react.default.createElement(_react.default.Fragment, null));
  }

}

exports.InputContainer = InputContainer;