"use strict";

var _interopRequireDefault = require("@babel/runtime/helpers/interopRequireDefault");

exports.__esModule = true;
exports.AsyncCall = exports.getTypeOperation = exports.initialStateAsyncCall = void 0;

var _immer = _interopRequireDefault(require("immer"));

var _rxjs = require("rxjs");

var _lang = require("../../../utils/lang");

var _model = require("../../common/model");

const initialStateAsyncCall = {
  params: {},
  status: {},
  errors: []
};
exports.initialStateAsyncCall = initialStateAsyncCall;

const getTypeOperation = processingType => {
  switch (processingType) {
    case _model.ProcessingType.CONCAT:
      return _rxjs.concatMap;

    case _model.ProcessingType.EXHAUST:
      return _rxjs.exhaustMap;

    case _model.ProcessingType.FLAT:
      return _rxjs.flatMap;

    case _model.ProcessingType.MERGE:
      return _rxjs.mergeMap;

    case _model.ProcessingType.SWITCH:
      return _rxjs.switchMap;

    default:
      return _rxjs.switchMap;
  }
};

exports.getTypeOperation = getTypeOperation;

class AsyncCall {
  constructor(executeFn$, config, setState, destroy$) {
    this.executeFn$ = executeFn$;
    this.config = config;
    this.setState = setState;
    this.destroy$ = destroy$;
    this.execute$ = new _rxjs.Subject();
    this.newState = initialStateAsyncCall;
    this.setData$Subject = new _rxjs.Subject();
    this.destroy$ = destroy$ ? destroy$ : new _rxjs.Subject();
    this.init = this.init.bind(this);
    this.getState = this.getState.bind(this);
    this.reload = this.reload.bind(this);
    this.initSetData = this.initSetData.bind(this);
    this.getStatus = this.getStatus.bind(this);
  }

  getStatus() {
    return this.newState.status.status;
  }

  initSetData() {
    this.setData$Subject.pipe((0, _rxjs.takeUntil)(this.destroy$), (0, _rxjs.tap)(setData => {
      this.setState(prevState => {
        const newState = (0, _immer.default)(draft => {
          const state = this.getState();
          debugger;
          draft.status.data = setData(state.status.data);
        })(prevState);
        this.newState = newState;
        return newState;
      });
    })).subscribe();
  }

  init() {
    this.initSetData();
    const processingStatus$ = new _rxjs.Subject();
    const config = this.config ? this.config : {};
    const params$ = config.source$ ? (0, _rxjs.merge)(this.config.source$, this.execute$) : this.execute$;
    const asyncOperation = getTypeOperation(config.processingType);
    (0, _rxjs.merge)(processingStatus$.pipe((0, _rxjs.mapTo)({
      status: "PROCESSING"
    })), params$.pipe((0, _rxjs.takeUntil)(this.destroy$), (0, _rxjs.debounceTime)(10), (0, _rxjs.tap)(params => {
      this.setState(prevState => {
        const newState = (0, _immer.default)(draft => {
          draft.params = params;
          draft.errors = [];
        })(prevState);
        this.newState = newState;
        return newState;
      });
      processingStatus$.next();
    }), asyncOperation(params => {
      const data$ = this.executeFn$ ? this.executeFn$(params) : (0, _rxjs.of)(params);
      const loadData$ = _lang.ModuleUtils.isObservable(data$) ? data$ : _lang.ModuleUtils.isPromise(data$) ? (0, _rxjs.from)(data$) : (0, _rxjs.of)(data$);
      const retryConfig = config.retry ? config.retry : {
        interval: 1000,
        maxRetryAttempts: 0
      };
      const maxRetryAttempts = retryConfig.maxRetryAttempts ? retryConfig.maxRetryAttempts : 0;
      const retryWhenStatus = retryConfig.retryWhenStatus ? retryConfig.retryWhenStatus : [];
      const notRetryWhenStatus = retryConfig.notRetryWhenStatus ? retryConfig.notRetryWhenStatus : [];
      return loadData$.pipe((0, _rxjs.retryWhen)(errors => errors.pipe((0, _rxjs.tap)(e => {
        this.setState(prevState => {
          const newState = (0, _immer.default)(draft => {
            draft.errors.push(e);
          })(prevState);
          this.newState = newState;
          return newState;
        });
      }), (0, _rxjs.concatMap)((e, i) => (0, _rxjs.iif)(() => {
        if (i < maxRetryAttempts) {
          const responseStatus = (e || {}).status || 0;
          const retry = !!retryWhenStatus.find(status => status === responseStatus);

          if (retry) {
            return false;
          }

          const notRetry = !!notRetryWhenStatus.find(status => status === responseStatus);

          if (notRetry) {
            return true;
          }

          if (retryConfig.retryWhen && retryConfig.retryWhen(e)) {
            return false;
          }

          if (retryConfig.noRetryWhen && retryConfig.noRetryWhen(e)) {
            return true;
          }

          return false;
        }

        return true;
      }, (0, _rxjs.throwError)(e), (0, _rxjs.of)(e).pipe((0, _rxjs.delay)((retryConfig.typeInterval === "LINEAR" && 1 || retryConfig.typeInterval === "EXPONENTIAL" && i) * retryConfig.interval)))))), (0, _rxjs.take)(1), (0, _rxjs.map)(data => {
        return {
          status: "SUCCESS",
          data
        };
      }), (0, _rxjs.catchError)(error => {
        return (0, _rxjs.of)({
          status: "ERROR",
          error
        });
      }));
    }))).pipe((0, _rxjs.takeUntil)(this.destroy$), (0, _rxjs.tap)(processingStatus => {
      this.setState(prevState => {
        const newState = (0, _immer.default)(draft => {
          draft.status.data = processingStatus.data;
          draft.status.error = processingStatus.error;
        })(prevState);
        this.newState = newState;
        return newState;
      });
    }), (0, _rxjs.tap)(processingStatus => {
      const params = this.newState.params;
      processingStatus.status === "PROCESSING" && config.onProcessing && config.onProcessing(params);
      processingStatus.status === "ERROR" && config.onError && config.onError(processingStatus.error, params);
      processingStatus.status === "SUCCESS" && config.onSuccess && config.onSuccess(processingStatus.data, params);
      this.setState(prevState => {
        const newState = (0, _immer.default)(draft => {
          draft.status.status = processingStatus.status;
        })(prevState);
        this.newState = newState;
        return newState;
      });
    })).subscribe();
  }

  getState() {
    return this.newState;
  }

  reload() {
    this.execute$.next(this.newState.params);
  }

  get execute() {
    return this.execute$.next.bind(this.execute$);
  }

  get setData() {
    return this.setData$Subject.next.bind(this.setData$Subject);
  }

}

exports.AsyncCall = AsyncCall;