"use strict";

var _interopRequireDefault = require("@babel/runtime/helpers/interopRequireDefault");

exports.__esModule = true;
exports.isEvent = exports.isReactNative = exports.FormArray = exports.FormGroup = exports.FormControl = exports.AbstractControl = exports.DISABLED = exports.PENDING = exports.INVALID = exports.VALID = void 0;

var _extends2 = _interopRequireDefault(require("@babel/runtime/helpers/extends"));

var _validators = require("./validators");

var _event_emitter = require("./core/event_emitter");

var _immer = _interopRequireDefault(require("immer"));

var property = _interopRequireWildcard(require("./utils/property"));

var _shared = require("./core/shared");

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function (nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

/**
 * @license
 * Copyright Google LLC All Rights Reserved.
 *
 * Use of this source code is governed by an MIT-style license that can be
 * found in the LICENSE file at https://angular.io/license
 */
const VALID = "VALID";
exports.VALID = VALID;
const INVALID = "INVALID";
exports.INVALID = INVALID;
const PENDING = "PENDING";
exports.PENDING = PENDING;
const DISABLED = "DISABLED";
exports.DISABLED = DISABLED;

function _find(control, path, delimiter) {
  if (path == null) return null;

  if (!Array.isArray(path)) {
    path = path.split(delimiter);
  }

  if (Array.isArray(path) && path.length === 0) return null;
  let controlToFind = control;
  path.forEach(name => {
    if (controlToFind instanceof FormGroup) {
      controlToFind = controlToFind.controls.hasOwnProperty(name) ? controlToFind.controls[name] : null;
    } else if (controlToFind instanceof FormArray) {
      controlToFind = controlToFind.at(name) || null;
    } else {
      controlToFind = null;
    }
  });
  return controlToFind;
}

function pickValidators(validatorOrOpts) {
  return (isOptionsObj(validatorOrOpts) ? validatorOrOpts.validators : validatorOrOpts) || null;
}

function coerceToValidator(validator) {
  return Array.isArray(validator) ? (0, _validators.composeValidators)(validator) : validator || null;
}

function pickAsyncValidators(asyncValidator, validatorOrOpts) {
  return (isOptionsObj(validatorOrOpts) ? validatorOrOpts.asyncValidators : asyncValidator) || null;
}

function coerceToAsyncValidator(asyncValidator) {
  return Array.isArray(asyncValidator) ? (0, _validators.composeAsyncValidators)(asyncValidator) : asyncValidator || null;
}

function isOptionsObj(validatorOrOpts) {
  return validatorOrOpts != null && !Array.isArray(validatorOrOpts) && typeof validatorOrOpts === "object";
}

class AbstractControl {
  constructor(validators, asyncValidators) {
    this._pendingDirty = void 0;
    this._hasOwnPendingAsyncValidator = false;
    this._pendingTouched = void 0;

    this._onCollectionChange = () => {};

    this._updateOn = void 0;
    this._parent = null;
    this._asyncValidationSubscription = void 0;
    this._composedValidatorFn = void 0;
    this._composedAsyncValidatorFn = void 0;
    this._rawValidators = void 0;
    this._rawAsyncValidators = void 0;
    this.value = void 0;
    this.status = void 0;
    this.errors = void 0;
    this.pristine = true;
    this.touched = false;
    this.valueChanges = void 0;
    this.statusChanges = void 0;
    this._onDisabledChange = [];
    this.nestedPath = void 0;
    this.path = null;

    this.dispatchStateFn = vale => {};

    this._rawValidators = validators;
    this._rawAsyncValidators = asyncValidators;
    this._composedValidatorFn = coerceToValidator(this._rawValidators);
    this._composedAsyncValidatorFn = coerceToAsyncValidator(this._rawAsyncValidators);
  }

  get validator() {
    return this._composedValidatorFn;
  }

  set validator(validatorFn) {
    this._rawValidators = this._composedValidatorFn = validatorFn;
  }

  get asyncValidator() {
    return this._composedAsyncValidatorFn;
  }

  set asyncValidator(asyncValidatorFn) {
    this._rawAsyncValidators = this._composedAsyncValidatorFn = asyncValidatorFn;
  }

  get parent() {
    return this._parent;
  }

  get valid() {
    return this.status === VALID;
  }

  get invalid() {
    return this.status === INVALID;
  }

  get pending() {
    return this.status == PENDING;
  }

  get disabled() {
    return this.status === DISABLED;
  }

  get enabled() {
    return this.status !== DISABLED;
  }

  get dirty() {
    return !this.pristine;
  }

  get untouched() {
    return !this.touched;
  }

  get updateOn() {
    return this._updateOn ? this._updateOn : this.parent ? this.parent.updateOn : "change";
  }

  setValidators(validators) {
    this._rawValidators = validators;
    this._composedValidatorFn = coerceToValidator(validators);
  }

  setAsyncValidators(validators) {
    this._rawAsyncValidators = validators;
    this._composedAsyncValidatorFn = coerceToAsyncValidator(validators);
  }

  addValidators(validators) {
    this.setValidators((0, _validators.addValidators)(validators, this._rawValidators));
  }

  addAsyncValidators(validators) {
    this.setAsyncValidators((0, _validators.addValidators)(validators, this._rawAsyncValidators));
  }

  removeValidators(validators) {
    this.setValidators((0, _validators.removeValidators)(validators, this._rawValidators));
  }

  removeAsyncValidators(validators) {
    this.setAsyncValidators((0, _validators.removeValidators)(validators, this._rawAsyncValidators));
  }

  hasValidator(validator) {
    return (0, _validators.hasValidator)(this._rawValidators, validator);
  }

  hasAsyncValidator(validator) {
    return (0, _validators.hasValidator)(this._rawAsyncValidators, validator);
  }

  clearValidators() {
    this.validator = null;
  }

  clearAsyncValidators() {
    this.asyncValidator = null;
  }

  markAsTouched(opts = {}) {
    this.touched = true;

    if (this._parent && !opts.onlySelf) {
      this._parent.markAsTouched(opts);
    }
  }

  markAllAsTouched() {
    this.markAsTouched({
      onlySelf: true
    });

    this._forEachChild(control => control.markAllAsTouched());
  }

  markAsUntouched(opts = {}) {
    this.touched = false;
    this._pendingTouched = false;

    this._forEachChild(control => {
      control.markAsUntouched({
        onlySelf: true
      });
    });

    if (this._parent && !opts.onlySelf) {
      this._parent._updateTouched(opts);
    }
  }

  markAsDirty(opts = {}) {
    this.pristine = false;

    if (this._parent && !opts.onlySelf) {
      this._parent.markAsDirty(opts);
    }
  }

  markAsPristine(opts = {}) {
    this.pristine = true;
    this._pendingDirty = false;

    this._forEachChild(control => {
      control.markAsPristine({
        onlySelf: true
      });
    });

    if (this._parent && !opts.onlySelf) {
      this._parent._updatePristine(opts);
    }
  }

  markAsPending(opts = {}) {
    this.status = PENDING;

    if (opts.emitEvent !== false) {
      this.statusChanges.emit(this.status);
    }

    if (this._parent && !opts.onlySelf) {
      this._parent.markAsPending(opts);
    }
  }

  disable(opts = {}) {
    const skipPristineCheck = this._parentMarkedDirty(opts.onlySelf);

    this.status = DISABLED;
    this.errors = null;

    this._forEachChild(control => {
      control.disable((0, _extends2.default)({}, opts, {
        onlySelf: true
      }));
    });

    this._updateValue();

    if (opts.emitEvent !== false) {
      this.valueChanges.emit(this.value);
      this.statusChanges.emit(this.status);
    }

    this._updateAncestors((0, _extends2.default)({}, opts, {
      skipPristineCheck
    }));

    this._onDisabledChange.forEach(changeFn => changeFn(true));
  }

  enable(opts = {}) {
    // If parent has been marked artificially dirty we don't want to re-calculate the
    // parent's dirtiness based on the children.
    const skipPristineCheck = this._parentMarkedDirty(opts.onlySelf);

    this.status = VALID;

    this._forEachChild(control => {
      control.enable((0, _extends2.default)({}, opts, {
        onlySelf: true
      }));
    });

    this.updateValueAndValidity({
      onlySelf: true,
      emitEvent: opts.emitEvent
    });

    this._updateAncestors((0, _extends2.default)({}, opts, {
      skipPristineCheck
    }));

    this._onDisabledChange.forEach(changeFn => changeFn(false));
  }

  _updateAncestors(opts) {
    if (this._parent && !opts.onlySelf) {
      this._parent.updateValueAndValidity(opts);

      if (!opts.skipPristineCheck) {
        this._parent._updatePristine();
      }

      this._parent._updateTouched();
    }
  }

  setParent(parent) {
    this._parent = parent;
  }

  updateValueAndValidity(opts = {}) {
    this._setInitialStatus();

    this._updateValue();

    if (this.enabled) {
      this._cancelExistingSubscription();

      this.errors = this._runValidator();
      this.status = this._calculateStatus();

      if (this.status === VALID || this.status === PENDING) {
        this._runAsyncValidator(opts.emitEvent);
      }
    }

    if (opts.emitEvent !== false) {
      this.valueChanges.emit(this.value);
      this.statusChanges.emit(this.status);
    }

    if (this._parent && !opts.onlySelf) {
      this._parent.updateValueAndValidity(opts);
    }
  }

  _updateTreeValidity(opts = {
    emitEvent: true
  }) {
    this._forEachChild(ctrl => ctrl._updateTreeValidity(opts));

    this.updateValueAndValidity({
      onlySelf: true,
      emitEvent: opts.emitEvent
    });
  }

  _setInitialStatus() {
    this.status = this._allControlsDisabled() ? DISABLED : VALID;
  }

  _runValidator() {
    return this.validator ? this.validator(this) : null;
  }

  _runAsyncValidator(emitEvent) {
    if (this.asyncValidator) {
      this.status = PENDING;
      this._hasOwnPendingAsyncValidator = true;
      const obs = (0, _validators.toObservable)(this.asyncValidator(this));
      this._asyncValidationSubscription = obs.subscribe(errors => {
        this._hasOwnPendingAsyncValidator = false;
        this.setErrors(errors, {
          emitEvent
        });
      });
    }
  }

  _cancelExistingSubscription() {
    if (this._asyncValidationSubscription) {
      this._asyncValidationSubscription.unsubscribe();

      this._hasOwnPendingAsyncValidator = false;
    }
  }

  setErrors(errors, opts = {}) {
    this.errors = errors;

    this._updateControlsErrors(opts.emitEvent !== false);
  }

  get(path) {
    return _find(this, path, ".");
  }

  getError(errorCode, path) {
    const control = path ? this.get(path) : this;
    return control && control.errors ? control.errors[errorCode] : null;
  }

  hasError(errorCode, path) {
    return !!this.getError(errorCode, path);
  }

  get root() {
    let x = this;

    while (x._parent) {
      x = x._parent;
    }

    return x;
  }
  /** @internal */


  _updateControlsErrors(emitEvent) {
    this.status = this._calculateStatus();

    if (emitEvent) {
      this.statusChanges.emit(this.status);
    }

    if (this._parent) {
      this._parent._updateControlsErrors(emitEvent);
    }
  }
  /** @internal */


  _initObservables() {
    this.valueChanges = new _event_emitter.EventEmitter2();
    this.statusChanges = new _event_emitter.EventEmitter2();
  }

  _calculateStatus() {
    if (this._allControlsDisabled()) return DISABLED;
    if (this.errors) return INVALID;
    if (this._hasOwnPendingAsyncValidator || this._anyControlsHaveStatus(PENDING)) return PENDING;
    if (this._anyControlsHaveStatus(INVALID)) return INVALID;
    return VALID;
  }
  /** @internal */


  /** @internal */
  _anyControlsHaveStatus(status) {
    return this._anyControls(control => control.status === status);
  }
  /** @internal */


  _anyControlsDirty() {
    return this._anyControls(control => control.dirty);
  }
  /** @internal */


  _anyControlsTouched() {
    return this._anyControls(control => control.touched);
  }
  /** @internal */


  _updatePristine(opts = {}) {
    this.pristine = !this._anyControlsDirty();

    if (this._parent && !opts.onlySelf) {
      this._parent._updatePristine(opts);
    }
  }
  /** @internal */


  _updateTouched(opts = {}) {
    this.touched = this._anyControlsTouched();

    if (this._parent && !opts.onlySelf) {
      this._parent._updateTouched(opts);
    }
  }
  /** @internal */


  /** @internal */
  _isBoxedValue(formState) {
    return typeof formState === "object" && formState !== null && Object.keys(formState).length === 2 && "value" in formState && "disabled" in formState;
  }
  /** @internal */


  _registerOnCollectionChange(fn) {
    this._onCollectionChange = fn;
  }
  /** @internal */


  _setUpdateStrategy(opts) {
    if (isOptionsObj(opts) && opts.updateOn != null) {
      this._updateOn = opts.updateOn;
    }
  }

  _parentMarkedDirty(onlySelf) {
    const parentDirty = this._parent && this._parent.dirty;
    return !onlySelf && !!parentDirty && !this._parent._anyControlsDirty();
  }

  getRoot() {
    if (this.parent) {
      return this.parent.getRoot();
    } else {
      return this;
    }
  }

  setState(value) {
    const root = this.getRoot();
    const dispatchStateFn = root.dispatchStateFn;
    const path = (root.nestedPath ? [root.nestedPath, ...this.getPath()] : this.getPath()).join(".");

    if (path === "") {
      dispatchStateFn(value);
    } else {
      const newState = (0, _immer.default)(draftState => {
        property.setValue(draftState, path, value);
      });
      dispatchStateFn(newState);
    }
  }

  getPath() {
    if (this.parent) return [...this.parent.getPath(), this.path];else return [];
  }

  getPathString() {
    return this.getPath().join(".");
  }

}

exports.AbstractControl = AbstractControl;

class FormControl extends AbstractControl {
  constructor(formState = null, validatorOrOpts, asyncValidator) {
    super(pickValidators(validatorOrOpts), pickAsyncValidators(asyncValidator, validatorOrOpts));
    this._onChange = [];
    this._pendingValue = void 0;
    this._pendingChange = void 0;

    this.onChange = event => {
      const value = getControlValue(event);
      const isDirty = value !== this.value;

      if (this.updateOn !== "change") {
        this._pendingValue = value;
        this._pendingChange = true;

        if (isDirty && !this._pendingDirty) {
          this._pendingDirty = true;
        }
      } else {
        if (isDirty && !this.dirty) {
          this.markAsDirty();
        }

        this.setValue(value);
      }
    };

    this.onBlur = () => {
      if (this.updateOn === "blur") {
        if (this._pendingDirty && !this.dirty) {
          this.markAsDirty();
        }

        if (!this.touched) {
          this.markAsTouched();
        }

        this.setValue(this._pendingValue, {
          emitEvent: false,
          onlySelf: true
        });
      } else if (this.updateOn === "submit") {
        this._pendingTouched = true;
      } else {
        const emitChangeToView = !this.touched;

        if (!this.touched) {
          this.markAsTouched();
          this.setValue(this.value);
        }

        if (emitChangeToView) {}
      }
    };

    this._applyFormState(formState);

    this._setUpdateStrategy(validatorOrOpts);

    this._initObservables();

    this.updateValueAndValidity({
      onlySelf: true,
      emitEvent: !!this.asyncValidator
    });
  }

  setValue(value, options = {}) {
    this.value = this._pendingValue = value;
    const root = this.getRoot();
    const tempPath = this.getPath().join(".");
    const path = root.nestedPath ? `${root.nestedPath}.${tempPath}` : tempPath;
    this.setState(value);

    if (this._onChange.length && options.emitModelToViewChange !== false) {
      this._onChange.forEach(changeFn => changeFn(this.value, options.emitViewToModelChange !== false));
    }

    this.updateValueAndValidity(options);
  }

  patchValue(value, options = {}) {
    this.setValue(value, options);
  }

  reset(formState = null, options = {}) {
    this._applyFormState(formState);

    this.markAsPristine(options);
    this.markAsUntouched(options);
    this.setValue(this.value, options);
    this._pendingChange = false;
  }

  _updateValue() {}

  _anyControls(condition) {
    return false;
  }

  _allControlsDisabled() {
    return this.disabled;
  }

  registerOnChange(fn) {
    this._onChange.push(fn);
  }

  _unregisterOnChange(fn) {
    (0, _shared.removeListItem)(this._onChange, fn);
  }

  registerOnDisabledChange(fn) {
    this._onDisabledChange.push(fn);
  }

  _unregisterOnDisabledChange(fn) {
    (0, _shared.removeListItem)(this._onDisabledChange, fn);
  }

  _forEachChild(cb) {}

  _syncPendingControls() {
    if (this.updateOn === "submit") {
      if (this._pendingDirty) this.markAsDirty();
      if (this._pendingTouched) this.markAsTouched();

      if (this._pendingChange) {
        this.setValue(this._pendingValue, {
          onlySelf: true,
          emitModelToViewChange: false
        });
        return true;
      }
    }

    return false;
  }

  _applyFormState(formState) {
    if (this._isBoxedValue(formState)) {
      this.value = this._pendingValue = formState.value;
      formState.disabled ? this.disable({
        onlySelf: true,
        emitEvent: false
      }) : this.enable({
        onlySelf: true,
        emitEvent: false
      });
    } else {
      this.value = this._pendingValue = formState;
    }
  }

}

exports.FormControl = FormControl;

class FormGroup extends AbstractControl {
  constructor(controls, validatorOrOpts, asyncValidator) {
    super(pickValidators(validatorOrOpts), pickAsyncValidators(asyncValidator, validatorOrOpts));
    this.controls = controls;

    this._initObservables();

    this._setUpdateStrategy(validatorOrOpts);

    this._setUpControls();

    this.updateValueAndValidity({
      onlySelf: true,
      emitEvent: !!this.asyncValidator
    });
  }

  registerControl(name, control) {
    if (this.controls[name]) return this.controls[name];
    this.controls[name] = control;
    control.setParent(this);

    control._registerOnCollectionChange(this._onCollectionChange);

    return control;
  }

  addControl(name, control, options = {}) {
    this.registerControl(name, control);
    this.updateValueAndValidity({
      emitEvent: options.emitEvent
    });

    this._onCollectionChange();
  }

  removeControl(name, options = {}) {
    if (this.controls[name]) this.controls[name]._registerOnCollectionChange(() => {});
    delete this.controls[name];
    this.updateValueAndValidity({
      emitEvent: options.emitEvent
    });

    this._onCollectionChange();
  }

  setControl(name, control, options = {}) {
    if (this.controls[name]) this.controls[name]._registerOnCollectionChange(() => {});
    delete this.controls[name];
    if (control) this.registerControl(name, control);
    this.updateValueAndValidity({
      emitEvent: options.emitEvent
    });

    this._onCollectionChange();
  }

  contains(controlName) {
    return this.controls.hasOwnProperty(controlName) && this.controls[controlName].enabled;
  }

  setValue(value, options = {}) {
    this._checkAllValuesPresent(value);

    Object.keys(value).forEach(name => {
      this._throwIfControlMissing(name);

      this.controls[name].setValue(value[name], {
        onlySelf: true,
        emitEvent: options.emitEvent
      });
    });
    this.updateValueAndValidity(options);
  }

  patchValue(value, options = {}) {
    if (value == null
    /* both `null` and `undefined` */
    ) return;
    Object.keys(value).forEach(name => {
      if (this.controls[name]) {
        this.controls[name].patchValue(value[name], {
          onlySelf: true,
          emitEvent: options.emitEvent
        });
      }
    });
    this.updateValueAndValidity(options);
    this.setState(value);
  }

  reset(value = {}, options = {}) {
    this._forEachChild((control, name) => {
      control.reset(value[name], {
        onlySelf: true,
        emitEvent: options.emitEvent
      });
    });

    this._updatePristine(options);

    this._updateTouched(options);

    this.updateValueAndValidity(options);
  }

  getRawValue() {
    return this._reduceChildren({}, (acc, control, name) => {
      acc[name] = control instanceof FormControl ? control.value : control.getRawValue();
      return acc;
    });
  }

  _syncPendingControls() {
    let subtreeUpdated = this._reduceChildren(false, (updated, child) => {
      return child._syncPendingControls() ? true : updated;
    });

    if (subtreeUpdated) this.updateValueAndValidity({
      onlySelf: true
    });
    return subtreeUpdated;
  }

  _throwIfControlMissing(name) {
    if (!Object.keys(this.controls).length) {
      throw new Error(`
         There are no form controls registered with this group yet. If you're using ngModel,
         you may want to check next tick (e.g. use setTimeout).
       `);
    }

    if (!this.controls[name]) {
      throw new Error(`Cannot find form control with name: ${name}.`);
    }
  }

  _forEachChild(cb) {
    Object.keys(this.controls).forEach(key => {
      const control = this.controls[key];
      control && cb(control, key);
    });
  }

  _setUpControls() {
    this._forEachChild((control, key) => {
      control.setParent(this);
      control.path = key;

      control._registerOnCollectionChange(this._onCollectionChange);
    });
  }

  _updateValue() {
    this.value = this._reduceValue();
  }

  _anyControls(condition) {
    for (const controlName of Object.keys(this.controls)) {
      const control = this.controls[controlName];

      if (this.contains(controlName) && condition(control)) {
        return true;
      }
    }

    return false;
  }

  _reduceValue() {
    return this._reduceChildren({}, (acc, control, name) => {
      if (control.enabled || this.disabled) {
        acc[name] = control.value;
      }

      return acc;
    });
  }
  /** @internal */


  _reduceChildren(initValue, fn) {
    let res = initValue;

    this._forEachChild((control, name) => {
      res = fn(res, control, name);
    });

    return res;
  }

  _allControlsDisabled() {
    for (const controlName of Object.keys(this.controls)) {
      if (this.controls[controlName].enabled) {
        return false;
      }
    }

    return Object.keys(this.controls).length > 0 || this.disabled;
  }
  /** @internal */


  _checkAllValuesPresent(value) {
    this._forEachChild((control, name) => {
      if (value[name] === undefined) {
        throw new Error(`Must supply a value for form control with name: '${name}'.`);
      }
    });
  }

}

exports.FormGroup = FormGroup;

class FormArray extends AbstractControl {
  constructor(controls, validatorOrOpts, asyncValidator) {
    super(pickValidators(validatorOrOpts), pickAsyncValidators(asyncValidator, validatorOrOpts));
    this.controls = controls;

    this._initObservables();

    this._setUpdateStrategy(validatorOrOpts);

    this._setUpControls();

    this.updateValueAndValidity({
      onlySelf: true,
      emitEvent: !!this.asyncValidator
    });
  }

  at(index) {
    return this.controls[index];
  }

  push(control, options = {}) {
    this.controls.push(control);
    control.path = this.controls.length - 1;
    const path = this.getPath().join(".") + "." + control.path;
    const newState = (0, _immer.default)(draftState => {
      property.setValue(draftState, path, control.value);
    });
    this.setState(newState);

    this._registerControl(control);

    this.updateValueAndValidity({
      emitEvent: options.emitEvent
    });

    this._onCollectionChange();
  }

  insert(index, control, options = {}) {
    this.controls.splice(index, 0, control);

    this._registerControl(control);

    this.updateValueAndValidity({
      emitEvent: options.emitEvent
    });
  }

  removeAt(index, options = {}) {
    if (this.controls[index]) this.controls[index]._registerOnCollectionChange(() => {});
    const path = this.getPath().join(".");
    const newState = (0, _immer.default)(draftState => {
      const controls = property.getValue(draftState, path);
      controls.splice(index, 1);
    });
    this.setState(newState);
    this.controls.forEach((control, index) => {
      control.path = index;
    });
    this.controls.splice(index, 1);
    this.updateValueAndValidity({
      emitEvent: options.emitEvent
    });
  }

  setControl(index, control, options = {}) {
    if (this.controls[index]) this.controls[index]._registerOnCollectionChange(() => {});
    this.controls.splice(index, 1);

    if (control) {
      this.controls.splice(index, 0, control);

      this._registerControl(control);
    }

    this.updateValueAndValidity({
      emitEvent: options.emitEvent
    });

    this._onCollectionChange();
  }

  get length() {
    return this.controls.length;
  }

  setValue(value, options = {}) {
    this._checkAllValuesPresent(value);

    value.forEach((newValue, index) => {
      this._throwIfControlMissing(index);

      this.at(index).setValue(newValue, {
        onlySelf: true,
        emitEvent: options.emitEvent
      });
    });
    this.updateValueAndValidity(options);
  }

  patchValue(value, options = {}) {
    if (value == null
    /* both `null` and `undefined` */
    ) return;
    value.forEach((newValue, index) => {
      if (this.at(index)) {
        this.at(index).patchValue(newValue, {
          onlySelf: true,
          emitEvent: options.emitEvent
        });
      }
    });
    this.updateValueAndValidity(options);
  }

  reset(value = [], options = {}) {
    this._forEachChild((control, index) => {
      control.reset(value[index], {
        onlySelf: true,
        emitEvent: options.emitEvent
      });
    });

    this._updatePristine(options);

    this._updateTouched(options);

    this.updateValueAndValidity(options);
  }

  getRawValue() {
    return this.controls.map(control => {
      return control instanceof FormControl ? control.value : control.getRawValue();
    });
  }

  clear(options = {}) {
    if (this.controls.length < 1) return;

    this._forEachChild(control => control._registerOnCollectionChange(() => {}));

    this.controls.splice(0);
    this.updateValueAndValidity({
      emitEvent: options.emitEvent
    });
  }

  _syncPendingControls() {
    let subtreeUpdated = this.controls.reduce((updated, child) => {
      return child._syncPendingControls() ? true : updated;
    }, false);
    if (subtreeUpdated) this.updateValueAndValidity({
      onlySelf: true
    });
    return subtreeUpdated;
  }

  _throwIfControlMissing(index) {
    if (!this.controls.length) {
      throw new Error(`
         There are no form controls registered with this array yet. If you're using ngModel,
         you may want to check next tick (e.g. use setTimeout).
       `);
    }

    if (!this.at(index)) {
      throw new Error(`Cannot find form control at index ${index}`);
    }
  }

  _forEachChild(cb) {
    this.controls.forEach((control, index) => {
      cb(control, index);
    });
  }

  _updateValue() {
    this.value = this.controls.filter(control => control.enabled || this.disabled).map(control => control.value);
  }

  _anyControls(condition) {
    return this.controls.some(control => control.enabled && condition(control));
  }

  _setUpControls() {
    this._forEachChild((control, index) => {
      this._registerControl(control);

      control.path = index;
    });
  }

  _checkAllValuesPresent(value) {
    this._forEachChild((control, i) => {
      if (value[i] === undefined) {
        throw new Error(`Must supply a value for form control at index: ${i}.`);
      }
    });
  }

  _allControlsDisabled() {
    for (const control of this.controls) {
      if (control.enabled) return false;
    }

    return this.controls.length > 0 || this.disabled;
  }

  _registerControl(control) {
    control.setParent(this);

    control._registerOnCollectionChange(this._onCollectionChange);
  }

}

exports.FormArray = FormArray;

function getControlValue(event) {
  if (isEvent(event)) {
    switch (event.target.type) {
      case "checkbox":
        return event.target.checked;

      case "select-multiple":
        if (event.target.options) {
          let options = event.target.options;
          var value = [];

          for (var i = 0, l = options.length; i < l; i++) {
            if (options[i].selected) {
              value.push(options[i].value);
            }
          }

          return value;
        }

        return event.target.value;

      case "number":
        return event.target.valueAsNumber;

      default:
        return isReactNative() ? event.nativeEvent.text : event.target.value;
    }
  }

  return event;
}

const isReactNative = () => typeof window !== "undefined" && window.navigator && window.navigator.product && window.navigator.product === "ReactNative";

exports.isReactNative = isReactNative;

const isEvent = candidate => !!(candidate && candidate.stopPropagation && candidate.preventDefault);

exports.isEvent = isEvent;